#!/usr/bin/env python3
#
#  This file is part of ProfileView.
#
#  ProfileView is free software: you can redistribute it and/or modify
#  it under the terms of the CeCILL 2.1 Licence
#
#  ProfileView is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  You should have received a copy of the Licence CeCILL 2.1 along
#  with ProfileView. If not, see <https://cecill.info/>.
#

import sys, os, io
import argparse
import re
import gzip, pickle
from collections import defaultdict

import weblogolib as weblogo
from BitVector import BitVector
from Bio import SeqIO

from pv_utils import *
from hh_utils import *


class hhalignment(object):

    def __init__(self):
        self.queryName  = ""
        self.targetName = ""
        self.matchStr = ""
        self.confidenceStr = ""
        self.probab = 0.0
        self.evalue = 0.0
        self.score  = 0.0
        self.aligned_cols = 0
        self.identities = 0.0
        self.similarity = 0.0
        self.sum_probs = 0.0
        self.template_neff = 0.0
        self.reset_query()
        self.reset_target()

    def reset_query(self):
        self.queryLen   = 0
        self.queryStart = -1
        self.queryEnd   = -1
        self.queryConsensus   = ""
        self.queryReference   = ""

    def reset_target(self):
        self.targetLen   = 0
        self.targetStart = -1
        self.targetEnd   = -1
        self.targetConsensus  = ""
        self.targetReference  = ""

    def __str__(self):
        return "Query={}; Start={}; End={}; Len={}; Consensus={}; Reference={}; Valid={}\n".format(self.queryName,self.queryStart,self.queryEnd,self.queryLen,len(self.queryConsensus)>0,len(self.queryReference)>0,self.is_query_valid()) \
                + "Target={}; Start={}; End={}; Len={}; Consensus={}; Reference={}; Valid={}\n".format(self.targetName,self.targetStart,self.targetEnd,self.targetLen,len(self.targetConsensus)>0,len(self.targetReference)>0,self.is_target_valid()) \
                + "Confidence={}; Valid={}".format(len(self.confidenceStr)>0,self.is_valid())

    def is_query_valid(self):
        return self.queryName != "" and self.queryLen > 0 and self.queryStart >= 0 and self.queryEnd <= self.queryLen and \
               self.queryStart <= self.queryEnd and self.queryReference != "" and len(self.queryConsensus) == len(self.queryReference)

    def is_target_valid(self):
        return self.targetName != "" and self.targetLen > 0 and self.targetStart >= 0 and self.targetEnd <= self.targetLen and \
               self.targetStart <= self.targetEnd and self.targetReference != "" and len(self.targetConsensus) == len(self.targetReference)

    def is_valid(self):
        return self.is_query_valid() and self.is_target_valid() and self.confidenceStr != "" and self.matchStr != ""

    def well_aligned_regions(self):
        qgaps  = 0
        tgaps  = 0
        lastend = 0
        regions = []
        for m in re.finditer(r"[9]+",self.confidenceStr):
            beg,end = m.span()
            qgaps += sum( self.queryReference[i] == '-' for i in range(lastend,beg) )
            qbeg = self.queryStart + beg - qgaps
            qend = qbeg + end - beg
            tgaps += sum( self.targetReference[i] == '-' for i in range(lastend,beg) )
            tbeg = self.targetStart + beg - tgaps
            tend = tbeg + end - beg
            regions.append( ((qbeg,qend),(tbeg,tend)) )
            lastend = end
        return regions


def read_alignment(fh):
    tseq = ""
    tseqIndex = 0
    qseqIndex = 0
    hha  = hhalignment()
    line = fh.readline()
    while line:
        line = line.rstrip()
        if line.startswith("Query "): # Query initialization
            if hha.is_valid():
                yield hha
                hha = hhalignment()
            hha.reset_query()
            hha.reset_target()
            hha.queryName = line.split()[1]
            hha.targetName = ""
        elif line.startswith(">"): # Target initialization
            if hha.is_valid():
                yield hha
                hha = hhalignment()
            hha.reset_query()
            hha.reset_target()
            hha.targetName = line.split()[0][1:]
            # read following line: 
            # Probab=<val>  E-value=<val>  Score=<val>  Aligned_cols=<val>  Identities=<val>%  Similarity=<val>  Sum_probs=<val>  Template_Neff=<val>
            values = [ p.split('=')[1] for p in fh.readline().split() ]
            hha.probab = float(values[0])
            hha.evalue = float(values[1])
            hha.score  = float(values[2])
            hha.aligned_cols = int(values[3])
            hha.identities = int(values[4][:-1]) # last character is '%', it must be removed before converting the string
            hha.similarity = float(values[5])
            hha.sum_probs = float(values[6])
            hha.template_neff = float(values[7])
        elif line.startswith("Q "): # Query alignment lines
            _,qname,qbeg,qseq,qend,qlen = line.split()
            if qname in ["ss_pred","ss_dssp"]:
                continue
            elif qname == "Consensus":
                hha.queryConsensus += qseq
                # following the line just read, there is the match string between query and target consensus
                qseqIndex = line.index(qseq)
                qseq_len = len(qseq)
                match_line = fh.readline().rstrip()
                match_str = match_line[qseqIndex:qseqIndex+qseq_len].ljust(qseq_len)
                hha.matchStr += match_str
            else:
                hha.queryStart = int(qbeg)-1 if hha.queryStart < 0 else hha.queryStart # Note: hha.queryStart is negative when uninitialized, update only when the first part of the aligned is found
                hha.queryEnd = int(qend) # last value found should be the maximum one
                hha.queryLen = int(qlen[1:-1]) # the value is specified between round brackets
                hha.queryReference += qseq
        elif line.startswith("T "): # Target alignment lines
            _,tname,tbeg,tseq,tend,tlen = line.split()
            if tname in ["ss_pred","ss_dssp"]:
                continue
            elif tname == "Consensus":
                hha.targetConsensus += tseq
            else:
                hha.targetStart = int(tbeg)-1 if hha.targetStart < 0 else hha.targetStart # Note: hha.queryStart is negative when uninitialized, update only when the first part of the aligned is found
                hha.targetEnd = int(tend) # last value found should be the maximum one
                hha.targetLen = int(tlen[1:-1]) # the value is specified between round brackets
                hha.targetReference += tseq
                tseqIndex = line.index(tseq)
        elif line.startswith("Confidence"):
            tseq_len = len(tseq)
            conf_line = line[tseqIndex:tseqIndex+tseq_len].ljust(tseq_len).replace(' ','-')
            hha.confidenceStr += conf_line
            tseqIndex = 0
        line = fh.readline()
    if hha.is_valid(): # output last one
        yield hha


def main( argv = None ):

    # parameter definition
    parser = argparse.ArgumentParser()
    parser.add_argument('--hhm-dict', dest='hhmDictFile', required=True, type=str, help='HHM dictionary .pgz file')
    parser.add_argument('--hhm-list', dest='hhmListFile', help='List of HHM models to consider')
    parser.add_argument('--hhr-dir', dest='hhrDir', required=True, type=str, help='directory path of profile-profile comparisons in hhr format')
    parser.add_argument('--fas-dir', dest='fasDir', required=True, type=str, help='directory path of fas sequences')
    parser.add_argument('--out-dir', dest='outDir', required=True, type=str, metavar="prefix", help='output directory')
    args = parser.parse_args()

    # parameter validation
    validParameters = True;
    if not os.path.isfile(args.hhmDictFile):
        validParameters = False
        print_error('HHM dictionary file "{}" does not exist'.format(args.hhmDictFile))
    if args.hhmListFile and not os.path.isfile(args.hhmListFile):
        validParameters = False
        print_error('HHM list file "{}" does not exist'.format(args.hhmListFile))
    elif args.hhrDir is not None and not os.path.isdir(args.hhrDir):
        validParameters = False
        print_error('--hhr-dir \"{}\" is not a directory'.format(args.hhrDir))
    elif args.fasDir is not None and not os.path.isdir(args.fasDir):
        validParameters = False
        print_error('--fas-dir \"{}\" is not a directory'.format(args.fasDir))
    if not validParameters:
        return 1

    print_status("loading HHM dictionary and list")
    hhmDict = {}
    with gzip.open( args.hhmDictFile, 'r') as f:
        hhmDict = pickle.load(f)
    hhmNameDict = {}
    if args.hhmListFile:
        with open(args.hhmListFile,'r') as f:
            for line in f:
                name = [ x.strip() for x in line.split(',') ]
                if len(name) > 0 and (not name[0].startswith('#')):
                    hhm_id = name[0]
                    hhm_name = name[1] if len(name) > 1 else hhm_id
                    hhmNameDict[hhm_id] = hhm_name
    else:
        hhmNameDict = { k:k for k in hhmDict }

    print_status("loading HHR alignment files")
    fileList = []
    if args.hhrDir is not None:
        hhrdir = args.hhrDir[:-1] if args.hhrDir.endswith("/") else args.hhrDir
        fileList = [ '{}/{}'.format(hhrdir,f) for f in os.listdir(hhrdir) if f.endswith('.hhr') and os.path.isfile('{}/{}'.format(hhrdir,f)) ]

    alnDict = defaultdict(list)
    for hhrFileName in fileList:
        with open(hhrFileName,'r') as hhrFile:
            for hhalign in read_alignment(hhrFile):
                alnDict[hhalign.queryName].append(hhalign)

    for qname in alnDict:
        alnDict[qname].sort( key=lambda x:x.evalue )

#    with open(args.outFile, 'w') as of:
#        for queryName in alnDict:
#            for hha in alnDict[queryName]:
#                outList = [ hha.queryName, hha.targetName, hha.probab, hha.evalue, hha.score, hha.aligned_cols, hha.identities, hha.similarity, hha.sum_probs, hha.template_neff ]
#                of.write('\t'.join((str(x) for x in outList)) + '\n')

    if not os.path.exists(args.outDir):
        os.mkdir(args.outDir)

    for queryName in alnDict:
        if queryName not in hhmNameDict:
            continue
        hhm = hhmDict[queryName]
        #queryConsensus = [ qc for i,qc in enumerate(hhm.consensus) ] # if hhm.conserved[i] ]
        #queryCovered = BitVector(size=len(hhm.master_seq))

        # output sequence logo restricted to conserved positions
        with open( "{}/{}.fas".format( args.fasDir, queryName ), 'r' ) as fas:
            fas_str = ""
            for rec in SeqIO.parse(fas,'fasta'):
                fas_str += ">{}\n{}\n".format(rec.id,''.join([ c for i,c in enumerate(rec.seq) if hhm.conserved[i] ]))
            loptions = weblogo.LogoOptions(
                    logo_title="{} ({})".format(hhmNameDict[queryName],queryName),
                    unit_name="bits",
                    yaxis_label="",
                    #show_yaxis=False,
                    color_scheme=weblogo.std_color_schemes["chemistry"],
                    annotate=[ i+1 for i,b in enumerate(hhm.conserved) if b ],
                    rotate_numbers=True,
                    stack_aspect_ratio=3,
                    stack_width=weblogo.std_sizes["large"],
                    stacks_per_line=hhm.conserved.count_bits(),
                    show_fineprint=False )
            ldata = weblogo.LogoData.from_seqs( weblogo.read_seq_data(io.StringIO(fas_str)) )
            lformat = weblogo.LogoFormat(ldata,loptions)
            with open( os.path.join(args.outDir, "{}.conserved.pdf".format(hhmNameDict[queryName])), 'wb' ) as ofh:
                ofh.write( weblogo.pdf_formatter(ldata,lformat) )

        with open( "{}/{}.fas".format( args.fasDir, queryName ), 'r' ) as fas:
            loptions = weblogo.LogoOptions(
                    logo_title="{} ({})".format(hhmNameDict[queryName],queryName),
                    unit_name="bits",
                    yaxis_label="",
                    #show_yaxis=False,
                    color_scheme=weblogo.std_color_schemes["chemistry"],
                    #rotate_numbers=True,
                    stack_aspect_ratio=3,
                    stack_width=weblogo.std_sizes["large"],
                    stacks_per_line=len(hhm.conserved),
                    show_fineprint=False )
            ldata = weblogo.LogoData.from_seqs( weblogo.read_seq_data(fas) )
            lformat = weblogo.LogoFormat(ldata,loptions)
            with open( os.path.join(args.outDir, "{}.full.pdf".format(hhmNameDict[queryName])), 'wb' ) as ofh:
                ofh.write( weblogo.pdf_formatter(ldata,lformat) )

        otherConserved = BitVector(size=len(hhm.master_seq)) # position conserved in any other template that aligns on a given query postion
        consDict = { qi:[] for qi in range(hhm.size()) if hhm.conserved[qi] }
        lessConsDict = { qi:[] for qi in range(hhm.size()) if hhm.conserved[qi] }
        for a in alnDict[queryName]:
            if a.targetName in hhmNameDict:
                qi = a.queryStart
                ti = a.targetStart
                t_hhm = hhmDict[a.targetName]
                for j,qc in enumerate(a.queryConsensus):
                    tc = a.targetConsensus[j]
                    if qc == '-': # skip gaps in query
                        ti += 1
                        continue
                    if tc != '-' and t_hhm.conserved[ti]:
                        otherConserved[qi] = True
                    if hhm.conserved[qi] and a.confidenceStr[j] in ['8','9'] and a.targetConsensus[j]==qc and qc.isupper(): # same conserved residue in target
                        consDict[qi].append(a.targetName)
                    elif hhm.conserved[qi] and a.confidenceStr[j] in ['8','9'] and a.matchStr[j]=='|': # TODO: admit also a.targetConsensus[j]==qc.lower() ???
                        lessConsDict[qi].append(a.targetName)
                    qi += 1
                    ti += (1 if tc != '-' else 0)
        # output sequence logo of important positions (either conserved in the current model or in any other one that aligns against it
        with open( "{}/{}.fas".format( args.fasDir, queryName ), 'r' ) as fas:
            fas_str = ""
            for rec in SeqIO.parse(fas,'fasta'):
                fas_str += ">{}\n{}\n".format(rec.id,''.join([ c for i,c in enumerate(rec.seq) if (hhm.conserved[i] or otherConserved[i]) ]))
            ann_list = []
            for i in range(len(rec.seq)):
                if hhm.conserved[i]:
                    ann_list.append(str(i+1))
                elif otherConserved[i]:
                    ann_list.append('')
            loptions = weblogo.LogoOptions(
                    logo_title="{} ({})".format(hhmNameDict[queryName],queryName),
                    unit_name="bits",
                    yaxis_label="",
                    #show_yaxis=False,
                    color_scheme=weblogo.std_color_schemes["chemistry"],
                    annotate=ann_list,
                    rotate_numbers=True,
                    stack_aspect_ratio=3,
                    stack_width=weblogo.std_sizes["large"],
                    stacks_per_line=len(ann_list),
                    show_fineprint=False )
            ldata = weblogo.LogoData.from_seqs( weblogo.read_seq_data(io.StringIO(fas_str)) )
            lformat = weblogo.LogoFormat(ldata,loptions)
            with open( os.path.join(args.outDir, "{}.any_conserved.pdf".format(hhmNameDict[queryName])), 'wb' ) as ofh:
                ofh.write( weblogo.pdf_formatter(ldata,lformat) )

        # output conserved positions for this model
        with open( os.path.join(args.outDir, "{}.cons".format(hhmNameDict[queryName])), 'w') as ofh:
            for qi in sorted(consDict):
                record = [ hhmNameDict[queryName], str(qi+1), hhm.consensus[qi] ] + [ hhmNameDict[targetName] for targetName in consDict[qi] ] + [ "*"+hhmNameDict[targetName] for targetName in lessConsDict[qi] ]
                ofh.write( '{}\n'.format('\t'.join(record)) )

    return 0


# Check if the program is not being imported
if __name__ == "__main__":
    sys.exit(main())

